;;;; This file contains implementation-independent tests for package local
;;;; nicknames, and is part of the Package-Local Nicknames CDR.
;;;;
;;;; If loading this file does not cause an error, the tests have passed succesfully.

;;;; Copyright (c) 2013 Nikodemus Siivola and Rudolf Schlatte
;;;;
;;;; Permission is hereby granted, free of charge, to any person obtaining
;;;; a copy of this software and associated documentation files (the
;;;; "Software"), to deal in the Software without restriction, including
;;;; without limitation the rights to use, copy, modify, merge, publish,
;;;; distribute, sublicense, and/or sell copies of the Software, and to
;;;; permit persons to whom the Software is furnished to do so, subject to
;;;; the following conditions:
;;;;
;;;; The above copyright notice and this permission notice shall be
;;;; included in all copies or substantial portions of the Software.
;;;;
;;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
;;;; EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
;;;; MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
;;;; IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
;;;; CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
;;;; TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
;;;; SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

(in-package :cl-user)

;;;; Setup

(eval-when (:compile-toplevel :load-toplevel :execute)

  ;; Step 1: figure out what is the local extension package.

  (defvar *package-local-nicknames-names*
    '("ADD-PACKAGE-LOCAL-NICKNAME"
      "PACKAGE-LOCALLY-NICKNAMED-BY-LIST"
      "PACKAGE-LOCAL-NICKNAMES"
      "REMOVE-PACKAGE-LOCAL-NICKNAME"))

  (defvar *package-local-nicknames-home-package*
    (let (package)
      (dolist (name *package-local-nicknames-names*)
        (let ((packages (remove-if (lambda (p)
                                     (or (not p) (eq p (find-package :keyword))))
                                   (mapcar #'symbol-package (find-all-symbols name)))))
          (cond ((not packages)
                 (warn "No package with symbol named ~A." name))
                ((cdr packages)
                 (warn "~@<Multiple symbols named ~A (in packages: ~{~A~^, ~}). Cannot ~
                      use for determining extension package for package local nicknames.~:@>"
                       name (mapcar #'package-name packages)))
                ((not package)
                 (setf package (car packages)))
                ((not (eq package (car packages)))
                 (error "~@<Both ~A and ~A package package-local-nickname symbols: can't ~
                       know for sure which is right.~:@>"
                        package (car packages))))))
      (or package
          (error "~@<Could not determine correct extension package for ~
                package-local-nicknames.~:@>"))))

  ;; Step 2: Make a package that re-exports only the interface we want
  ;; to test.

  (let ((*package* (or (find-package :package-local-nicknames-cdr-ext)
                       (make-package :package-local-nicknames-cdr-ext))))
    (use-package *package-local-nicknames-home-package*)
    (export (mapcar (lambda (name)
                      (find-symbol name *package-local-nicknames-home-package*))
                    *package-local-nicknames-names*))))

;;;; Actual tests.

#-package-local-nicknames
(error ":PACKAGE-LOCAL-NICKNAMES not in *FEATURES*")

(defpackage :package-local-nicknames-cdr-tests
  (:use :cl :package-local-nicknames-cdr-ext))

(in-package :package-local-nicknames-cdr-tests)

(defmacro is ((test actual expected) &optional (name "Test"))
  (let ((actual-n (gensym "ACTUAL"))
        (expected-n (gensym "EXPECTED")))
    `(let ((,actual-n ,actual)
           (,expected-n ,expected))
       (unless (,test ,expected-n ,actual-n)
         (error "~@<~A failed: (~S ~S ~S) not true.~:@>"
                ,name ',test ,expected-n ,actual-n)))))

(flet ((clean (name)
         (when (find-package name)
           (delete-package name))))
  (clean :package-local-nicknames-cdr-tester)
  (clean :package-local-nicknames-cdr-test-aux)
  (clean :package-local-nicknames-cdr-test-wrong)
  (clean :package-local-nicknames-cdr-test-right))

(defpackage :package-local-nicknames-cdr-test-aux
  (:use)
  (:export "CONS"))

(defpackage :package-local-nicknames-cdr-test-extra
  (:use)
  (:export "CONS"))

(defpackage :package-local-nicknames-cdr-test-wrong
  (:use)
  (:export "X"))

(defpackage :package-local-nicknames-cdr-test-right
  (:use)
  (:export "X"))

(is (eq (eval `(defpackage :package-local-nicknames-cdr-tester
                 (:local-nicknames (:l :cl)
                                   (:x :package-local-nicknames-cdr-ext)
                                   (:t :package-local-nicknames-cdr-test-aux)
                                   (:package-local-nicknames-cdr-test-wrong
                                    :package-local-nicknames-cdr-test-right)
                                   (:package-local-nicknames-cdr-test-right
                                    :package-local-nicknames-cdr-test-wrong))))
        (find-package :package-local-nicknames-cdr-tester))
    "DEFPACKAGE accepts :LOCAL-NICKNAMES")

(let ((alist (package-local-nicknames :package-local-nicknames-cdr-tester)))
  (is (equal (cons "L" (find-package :cl))
             (assoc "L" alist :test #'string=))
      "DEFPACKAGE sets up local nicknames alist (1)")
  (is (equal (cons "X" (find-package :package-local-nicknames-cdr-ext))
             (assoc "X" alist :test #'string=))
      "DEFPACKAGE sets up local nicknames alist (2)")
  (is (eql 5 (length alist))
      "DEFPACKAGE sets up local nicknames alist (3)"))

(is (equal (let ((*package* (find-package :package-local-nicknames-cdr-tester)))
             (list (find-package :l)
                   (find-package :x)
                   (find-package :t)
                   (find-package :package-local-nicknames-cdr-test-right)
                   (find-package :package-local-nicknames-cdr-test-wrong)))
           (list (find-package :cl)
                 (find-package :package-local-nicknames-cdr-ext)
                 (find-package :package-local-nicknames-cdr-test-aux)
                 (find-package :package-local-nicknames-cdr-test-wrong)
                 (find-package :package-local-nicknames-cdr-test-right)))
    "Local nicknames effect FIND-PACKAGE after DEFPACKAGE.")

(is (equal (let ((*package* (find-package :package-local-nicknames-cdr-tester)))
             (list (find-symbol "CONS" :l)
                   (find-symbol "PACKAGE-LOCAL-NICKNAMES" :x)
                   (find-symbol "CONS" :t)
                   (find-symbol "X" :package-local-nicknames-cdr-test-wrong)
                   (find-symbol "X" :package-local-nicknames-cdr-test-right)))
           (list 'cl:cons
                 'package-local-nicknames-cdr-ext:package-local-nicknames
                 'package-local-nicknames-cdr-test-aux:cons
                 'package-local-nicknames-cdr-test-right:x
                 'package-local-nicknames-cdr-test-wrong:x))
    "Local nicknames effect FIND-SYMBOL after DEFPACKAGE.")

(is (equal (let ((*package* (find-package :package-local-nicknames-cdr-tester)))
             (list (intern "CONS" :l)
                   (intern "PACKAGE-LOCAL-NICKNAMES" :x)
                   (intern "CONS" :t)
                   (find-symbol "X" :package-local-nicknames-cdr-test-wrong)
                   (find-symbol "X" :package-local-nicknames-cdr-test-right)
                   (symbol-package (intern "ENTIRELY-NEW-SYMBOL" :t))))
           (list 'cl:cons
                 'package-local-nicknames-cdr-ext:package-local-nicknames
                 'package-local-nicknames-cdr-test-aux:cons
                 'package-local-nicknames-cdr-test-right:x
                 'package-local-nicknames-cdr-test-wrong:x
                 (find-package :package-local-nicknames-cdr-test-aux)))
    "Local nicknames effect INTERN after DEFPACKAGE.")

(is (equal (let ((*package* (find-package :package-local-nicknames-cdr-tester)))
             (list (read-from-string "L:CONS")
                   (read-from-string "X:PACKAGE-LOCAL-NICKNAMES")
                   (read-from-string "T:CONS")
                   (read-from-string "PACKAGE-LOCAL-NICKNAMES-CDR-TEST-WRONG:X")
                   (read-from-string "PACKAGE-LOCAL-NICKNAMES-CDR-TEST-RIGHT:X")
                   (symbol-package (read-from-string "T::ENTIRELY-NEW-SYMBOL2"))))
           (list 'cl:cons
                 'package-local-nicknames-cdr-ext:package-local-nicknames
                 'package-local-nicknames-cdr-test-aux:cons
                 'package-local-nicknames-cdr-test-right:x
                 'package-local-nicknames-cdr-test-wrong:x
                 (find-package :package-local-nicknames-cdr-test-aux)))
    "Local nicknames effect READ-FROM-STRING after DEFPACKAGE.")

(with-standard-io-syntax
  (let ((*package* (find-package :package-local-nicknames-cdr-tester)))
    (is (string= "L:CONS" (prin1-to-string (read-from-string "L:CONS")))
        "Print-read consistency (1)")
    (is (string= "PACKAGE-LOCAL-NICKNAMES-CDR-TEST-WRONG:X"
                 (prin1-to-string
                  (read-from-string "PACKAGE-LOCAL-NICKNAMES-CDR-TEST-WRONG:X")))
        "Print-read consistency (2)")
    (is (string= "PACKAGE-LOCAL-NICKNAMES-CDR-TEST-RIGHT:X"
                 (prin1-to-string
                  (read-from-string "PACKAGE-LOCAL-NICKNAMES-CDR-TEST-RIGHT:X")))
        "Print-read consistency (3)")))

(let ((global (find-package :u))
      (*package* (find-package :package-local-nicknames-cdr-tester)))
  (is (eq *package* (add-package-local-nickname :u :cl-user))
      "ADD-PACKAGE-LOCAL-NICKNAME works")
  (is (eq (find-package :u) (find-package :cl-user))
      "ADD-PACKAGE-LOCAL-NICKNAME takes effect")
  (is (eq *package* (add-package-local-nickname :u :cl-user))
      "ADD-PACKAGE-LOCAL-NICKNAME can be repeated")
  (is (eq :oops
          (handler-case (add-package-local-nickname :u :cl)
            (error () :oops)))
      "ADD-PACKAGE-LOCAL-NICKNAME with same name for different packages signals an error.")
  (is (eq t (when (remove-package-local-nickname :u) t))
      "REMOVE-PACKAGE-LOCAL-NICKNAME works")
  (is (eq global (find-package :u))
      "REMOVE-PACKAGE-LOCAL-NICKNAME takes effect")
  (is (eq nil (remove-package-local-nickname :u))
      "REMOVE-PACKAGE-LOCAL-NICKNAME can be repeted"))

;;; XXX: (1) Is this allowed. (2) Are both U1::KLASLDKJAS and U2::KLASLDKJAS ok?
(let ((*package* (find-package :package-local-nicknames-cdr-tester)))
  (add-package-local-nickname :u1 :cl-user)
  (add-package-local-nickname :u2 :cl-user)
  (prin1-to-string (read-from-string "U1::KLASLDKJAS")))

(defmacro with-tmp-packages (bindings &body body)
  `(let ,(mapcar #'car bindings)
     (unwind-protect
          (progn
            (setf ,@(apply #'append bindings))
            ,@body)
       ,@(mapcar (lambda (p)
                   `(when ,p (delete-package ,p)))
                 (mapcar #'car bindings)))))

(with-tmp-packages ((p1 (make-package "LOCALLY-NICKNAMES-OTHERS"))
                    (p2 (make-package "LOCALLY-NICKNAMED-BY-OTHERS")))
  (add-package-local-nickname :foo p2 p1)
  (is (equal (list p1) (package-locally-nicknamed-by-list p2))
      "DELETE-PACKAGE (1) interaction precondition")
  (delete-package p1)
  (is (eq nil (package-locally-nicknamed-by-list p2))
      "DELETE-PACKAGE (1) interaction postcondition"))

(with-tmp-packages ((p1 (make-package "LOCALLY-NICKNAMES-OTHERS"))
                    (p2 (make-package "LOCALLY-NICKNAMED-BY-OTHERS")))
  (add-package-local-nickname :foo p2 p1)
  (is (equal (acons "FOO" (find-package p2) nil) (package-local-nicknames p1))
      "DELETE-PACKAGE (2) interaction precondition")
  (delete-package p2)
  (is (eq nil (package-local-nicknames p1))
      "DELETE-PACKAGE (2) interaction postcondition"))

(with-tmp-packages ((p1 (make-package "OWN-NAME-AS-NICKNAME1"))
                    (p2 (make-package "OWN-NAME-AS-NICKNAME2")))
  (is (eq :oops
          (handler-case
              (add-package-local-nickname :own-name-as-nickname1 p2 p1)
            (error () :oops)))
      "Own name as local nickname (1)")
  (handler-bind ((error #'continue))
    (add-package-local-nickname :own-name-as-nickname1 p2 p1))
  (is (eq (intern "FOO" p2)
          (let ((*package* p1))
            (intern "FOO" :own-name-as-nickname1)))
      "Own name as local nickname (2)"))

(with-tmp-packages ((p1 (make-package "OWN-NICKNAME-AS-NICKNAME1"
                                      :nicknames '("OWN-NICKNAME")))
                    (p2 (make-package "OWN-NICKNAME-AS-NICKNAME2")))
  (is (eq :oops
          (handler-case
              (add-package-local-nickname :own-nickname p2 p1)
            (error () :oops)))
      "Own nickname as local nickname (1)")
  (handler-bind ((error #'continue))
    (add-package-local-nickname :own-nickname p2 p1))
  (is (eq (intern "FOO" p2)
          (let ((*package* p1))
            (intern "FOO" :own-nickname)))
      "Own nickname as local nickname (2)"))
